unit LaunchU;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TForm1 = class(TForm)
    dlgOpen: TOpenDialog;
    lstWindows: TListBox;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  public
    Wnd: HWnd;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure DescribeWnd(Wnd: HWnd; List: TStrings);
const
  FlagStr: array[Boolean] of String = ('not ', '');
var
  CString: array[0..Max_Path] of Char;
  PID, TID: DWord;
  Rect: TRect;
begin
  with List do
  begin
    BeginUpdate;
    try
      //Check the window is valid
      if not IsWindow(Wnd) then
      begin
        Add('Window is not valid');
        Abort
      end;
      Add(Format('Window handle = $%x', [Wnd]));
      //Get window caption
      Win32Check(Bool(GetWindowText(Wnd, CString, SizeOf(CString))));
      Add(Format('Window caption is "%s"', [CString]));
      //get window dimensions
      Win32Check(GetWindowRect(Wnd, Rect));
      with Rect do
        Add(Format('Window co-ordinates are (%d,%d) - (%d,%d)',
          [Left, Top, Right, Bottom]));
      Add(Format('Window is from app with instance handle of $%x',
        [GetWindowLong(Wnd, gwl_HInstance)]));
      //Get thread id & process id of owning app
      TID := GetWindowThreadProcessID(Wnd, @PID);
      Add(Format('Window is from app with process identifier of $%x', [PID]));
      Add(Format('Window is from app with primary thread identifier of $%x', [TID]));
      Add(Format('Window has ID of $%x', [GetWindowLong(Wnd, gwl_ID)]));
      Add(Format('The window is %svisible', [FlagStr[IsWindowVisible(Wnd)]]));
      Add(Format('The window is %sa Unicode window', [FlagStr[IsWindowUnicode(Wnd)]]));
      Add(Format('The window is %senabled', [FlagStr[IsWindowEnabled(Wnd)]]));
      Application.Icon.Handle := GetClassLong(Wnd, gcl_HIcon);
      Add('The window''s icon has been set as this application''s icon');
      Add('The window''s caption bar should be flashing')
    finally
      EndUpdate
    end
  end
end;

function EnumFunc(Wnd: HWnd; PWnd: PHandle): Bool; stdcall;
begin
  //Terminate the enumeration
  Result := False;
  //Return the window handle
  PWnd^ := Wnd
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  CString: array[0..Max_Path] of Char;
  SI: TStartupInfo;
  PI: TProcessInformation;
begin
  //Set open dialog to look in Windows directory
  GetWindowsDirectory(CString, SizeOf(CString));
  dlgOpen.InitialDir := CString;
  if dlgOpen.Execute then
  begin
    //Launch chosen app
    GetStartupInfo(SI);
    Win32Check(CreateProcess(nil, PChar(dlgOpen.FileName),
      nil, nil, False, 0, nil, nil, SI, PI));
    //Wait for main window to initialise
    WaitForInputIdle(PI.hProcess, Infinite);
    //Loop thru all app's windows
    EnumThreadWindows(PI.dwThreadId, @EnumFunc, Longint(@Wnd));
    //Fill listbox with info on the main window
    DescribeWnd(Wnd, lstWindows.Items);
  end
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  //Flash the window's caption bar
  if IsWindow(Wnd) then
    FlashWindow(Wnd, True)
end;

end.
